#!/usr/bin/python

# This code is original from jsmin by Douglas Crockford, it was translated to
# Python by Baruch Even. The original code had the following copyright and
# license.
#
# /* jsmin.c
#    2007-01-08
#
# Copyright (c) 2002 Douglas Crockford  (www.crockford.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# The Software shall be used for Good, not Evil.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# */

from StringIO import StringIO

def about():
    print "Compressing javascript with jsmin backend"

def compress(js):
    ins = StringIO(js)
    outs = StringIO()
    JavascriptMinify().minify(ins, outs)
    str = outs.getvalue()
    if len(str) > 0 and str[0] == '\n':
        str = str[1:]
    return str

def isAlphanum(c):
    """return true if the character is a letter, digit, underscore,
           dollar sign, or non-ASCII character.
    """
    return ((c >= 'a' and c <= 'z') or (c >= '0' and c <= '9') or
            (c >= 'A' and c <= 'Z') or c == '_' or c == '$' or c == '\\' or (c is not None and ord(c) > 126));

class UnterminatedComment(Exception):
    pass

class UnterminatedStringLiteral(Exception):
    pass

class UnterminatedRegularExpression(Exception):
    pass

class JavascriptMinify(object):

    def _outA(self):
        self.outstream.write(self.theA)
    def _outB(self):
        self.outstream.write(self.theB)

    def _get(self):
        """return the next character from stdin. Watch out for lookahead. If
           the character is a control character, translate it to a space or
           linefeed.
        """
        c = self.theLookahead
        self.theLookahead = None
        if c == None:
            c = self.instream.read(1)
        if c >= ' ' or c == '\n':
            return c
        if c == '': # EOF
            return '\000'
        if c == '\r':
            return '\n'
        return ' '

    def _peek(self):
        self.theLookahead = self._get()
        return self.theLookahead

    def _next(self):
        """get the next character, excluding comments. peek() is used to see
           if a '/' is followed by a '/' or '*'.
        """
        c = self._get()
        if c == '/':
            p = self._peek()
            if p == '/':
                c = self._get()
                while c > '\n':
                    c = self._get()
                return c
            if p == '*':
                c = self._get()
                while 1:
                    c = self._get()
                    if c == '*':
                        if self._peek() == '/':
                            self._get()
                            return ' '
                    if c == '\000':
                        raise UnterminatedComment()

        return c

    def _action(self, action):
        """do something! What you do is determined by the argument:
           1   Output A. Copy B to A. Get the next B.
           2   Copy B to A. Get the next B. (Delete A).
           3   Get the next B. (Delete B).
           action treats a string as a single character. Wow!
           action recognizes a regular expression if it is preceded by ( or , or =.
        """
        if action <= 1:
            self._outA()

        if action <= 2:
            self.theA = self.theB
            if self.theA == "'" or self.theA == '"':
                while 1:
                    self._outA()
                    self.theA = self._get()
                    if self.theA == self.theB:
                        break
                    if self.theA <= '\n':
                        raise UnterminatedStringLiteral()
                    if self.theA == '\\':
                        self._outA()
                        self.theA = self._get()


        if action <= 3:
            self.theB = self._next()
            if self.theB == '/' and (self.theA == '(' or self.theA == ',' or
                                     self.theA == '=' or self.theA == ':' or
                                     self.theA == '[' or self.theA == '?' or
                                     self.theA == '!' or self.theA == '&' or
                                     self.theA == '|'):
                self._outA()
                self._outB()
                while 1:
                    self.theA = self._get()
                    if self.theA == '/':
                        break
                    elif self.theA == '\\':
                        self._outA()
                        self.theA = self._get()
                    elif self.theA <= '\n':
                        raise UnterminatedRegularExpression()
                    self._outA()
                self.theB = self._next()


    def _jsmin(self):
        """Copy the input to the output, deleting the characters which are
           insignificant to JavaScript. Comments will be removed. Tabs will be
           replaced with spaces. Carriage returns will be replaced with linefeeds.
           Most spaces and linefeeds will be removed.
        """
        self.theA = '\n'
        self._action(3)

        while self.theA != '\000':
            if self.theA == ' ':
                if isAlphanum(self.theB):
                    self._action(1)
                else:
                    self._action(2)
            elif self.theA == '\n':
                if self.theB in ['{', '[', '(', '+', '-']:
                    self._action(1)
                elif self.theB == ' ':
                    self._action(3)
                else:
                    if isAlphanum(self.theB):
                        self._action(1)
                    else:
                        self._action(2)
            else:
                if self.theB == ' ':
                    if isAlphanum(self.theA):
                        self._action(1)
                    else:
                        self._action(3)
                elif self.theB == '\n':
                    if self.theA in ['}', ']', ')', '+', '-', '"', '\'']:
                        self._action(1)
                    else:
                        if isAlphanum(self.theA):
                            self._action(1)
                        else:
                            self._action(3)
                else:
                    self._action(1)

    def minify(self, instream, outstream):
        self.instream = instream
        self.outstream = outstream
        self.theA = None
        self.thaB = None
        self.theLookahead = None

        self._jsmin()
        self.instream.close()

if __name__ == '__main__':
    import sys
    jsm = JavascriptMinify()
    jsm.minify(sys.stdin, sys.stdout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2008, 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# Code taked from Friendpaste under license:
# Apache 2 License
#
# 2008 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.
#

import re

__all__ = ['CSSParser']

re_selector = re.compile("/\./")
re_comments = re.compile("(\/\*).*?(\*\/)")
re_sep = re.compile(':\s*')
re_line1 = re.compile('\n')
re_line = re.compile('(\n)')
re_comma = re.compile(',')
re_comma2 = re.compile(',  ')


def strip_space(string):
    """  strip space after :, remove newlines, replace 
    multiple spaces with only one space, remove comments """
    if isinstance(string, basestring):
        string = re_line1.sub('', re_sep.sub(':', string))
        string = re_comments.sub('', string.strip())
    return string
    
def strip_selector_space(string):
    """ remove newlines, insert space after comma, replace 
    two spaces with one space after comma """
    if isinstance(string, basestring):
        string = re_comma2.sub(', ', re_comma.sub(', ', re_line.sub('', string)))
    return string
    
class CSSParser(object):
    
    def __init__(self, css_string, options=None):
        options = options or {}
        self.namespace = options.get('namespace', '')
        self.raw_data = css_string
        self.css_output = ''
        
        self._compress(self.raw_data)
        
    def __str__(self):
        return self.css_output
        
    def parse(self, data):
        data = data or self.raw_data
        css_out = []
        data = strip_space(data.strip())
        for index, assignements in enumerate(data.split('}')):
            try:
                tags, styles = [a.strip() for a in assignements.strip().split('{')]
                rules = []
                if styles:
                    if self.namespace:
                        tags = strip_selector_space(tags)
                        tags = re_selector.sub(self.namespace, tags)
                        rules = []
                    for key_val_pair in styles.split(';'):
                        try:
                            key, value =  [a.strip() for a in key_val_pair.split(':')]
                            rules.append("%s:%s;" % (key, value))
                        except ValueError:
                            continue
                    css_out.append({
                        'tags': tags,
                        'rules': ''.join(rules),
                        'idx': index 
                    })
            except ValueError:
                continue
        css_out.sort(lambda a,b: cmp(a['idx'], b['idx']))
        return css_out
        
    def _compress(self, data):
        self.css_output = '';

        for line in self.parse(data):
            self.css_output += "%s {%s}\n" % (line['tags'], line['rules'])    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.
#

import glob
import httplib
import os
import re
import shutil
import socket
import sys
import time
import urllib

# python 2.6
try:
    import json 
except ImportError:
    import simplejson as json 

# backport os.path.relpath if python < 2.6
try:
    import os.path.relpath as _relpath
except ImportError:
    def _relpath(path, start=os.curdir):
        if not path:
            raise ValueError("no path specified")
        
        start_list = os.path.abspath(start).split("/")
        path_list = os.path.abspath(path).split("/")

        # Work out how much of the filepath is shared by start and path.
        i = len(os.path.commonprefix([start_list, path_list]))

        rel_list = ['..'] * (len(start_list)-i) + path_list[i:]
        if not rel_list:
            return os.curdir
        return os.path.join(*rel_list)
        
try:#python 2.6, use subprocess
    from subprocess import *
    def _popen3(cmd, mode='t', bufsize=0):
        p = Popen(cmd, shell=True, bufsize=bufsize,
            stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
        return (p.stdin, p.stdout, p.stderr)
except ImportError:
    def _popen3(cmd, mode='t', bufsize=0):
        return os.popen3(cmd, mode, bufsize)
        


import httplib2
from couchdb import Server, ResourceNotFound

from couchapp.utils import _md5, to_bytestring
from couchapp.utils import *
from couchapp.utils.css_parser import CSSParser

__all__ = ['DEFAULT_SERVER_URI', 'FileManager']

DEFAULT_SERVER_URI = 'http://127.0.0.1:5984/'

# should be changes to jchris repo.
COUCHAPP_VENDOR_URL = 'git://github.com/benoitc/couchapp.git'
COUCHAPP_VENDOR_SCM = 'git'

external_dir = os.path.join(os.path.dirname(__file__), '_external')
VENDOR_HANDLERS = {
    'git': os.path.join(external_dir, 'git.sh')
}

def _server(server_uri):
    if "@" in server_uri:
        http = httplib2.Http()
        username, password, server_uri = parse_auth(server_uri) 
        couchdb_server = Server(server_uri)
        http.add_credentials(username, password)
        couchdb_server.resource.http = http
    else:
        couchdb_server = Server(server_uri)
    return couchdb_server

def get_userconf():
    """ return user conf """
    
    # this should work on windows too
    homedir = os.path.expanduser('~')
    user_conffile = os.path.join(homedir, ".couchapprc")
    if os.path.isfile(user_conffile):
        try:
            return read_json(user_conffile)
        except:
            pass
    return {}
    
def _vendor_handlers():
    user_conf = get_userconf()
    vendor_handlers = VENDOR_HANDLERS
    if "vendor_handlers" in user_conf:
        try:
            vendor_handler.update(user_conf['vendor_handlers'])
        except ValueError:
            pass
    return vendor_handlers

class FileManager(object):
    
    def __init__(self, dbstring, app_dir='.'):
        self.app_dir = app_dir 

        # load conf
        self.load_metadata(app_dir)
       
        if not dbstring or not "/" in dbstring:
            env = self.conf.get('env', {})
            if dbstring:
                if dbstring in env:
                    db_env = env[dbstring]['db']
                else: 
                    db_env = "%s/%s" % (DEFAULT_SERVER_URI, dbstring)
            else: 
                if 'default' in env:
                    db_env = env['default']['db']
                else:
                    raise ValueError("database isn't specified")

            if isinstance(db_env, basestring):
                self.db_url = [db_env]
            else:
                self.db_url = db_env
        else:
            self.db_url = [dbstring]

        db = []
        for s in self.db_url:
            server_uri, db_name, docid = parse_uri(s)
 
            couchdb_server = _server(server_uri)
            
            # create dbs if it don't exist
            try:
                _db = couchdb_server.create(db_name)
            except: # db already exist
                _db = couchdb_server[db_name]
            db.append(_db)
        self.db = db
        
        self.conf = get_userconf()
        
    
    @classmethod
    def generate_app(cls, app_dir):
        """Generates a CouchApp in app_dir"""
        template_paths = ['app-template', '../../app-template']
        vendor_paths = ['vendor', '../../vendor']
        
        for path in template_paths:
            template_dir = os.path.normpath(os.path.join(
                os.path.dirname(__file__), path))
            if os.path.isdir(template_dir): break
            
        for path in vendor_paths:
            vendor_dir = os.path.normpath(os.path.join(
                os.path.dirname(__file__), path))
            if os.path.isdir(vendor_dir): break
        
        try:
            shutil.copytree(template_dir, app_dir)
        except OSError, e:
            errno, message = e
            print >>sys.stderr, "Can't create a CouchApp in %s: %s" % (
                    app_dir, message)
            return
                          
        if vendor_dir:
            vendor_path = os.path.join(app_dir, 'vendor')
            try:
                shutil.copytree(vendor_dir, vendor_path)
            except OSError, e:
                errno, message = e
                print >>sys.stderr, "Can't create a CouchApp, bad vendor dir: %s" % message
                return

        cls.init(app_dir)
    
    @classmethod
    def init(cls, app_dir, db_url=''):
        """Initializes the .couchapprc, usually called after generate"""
        if not os.path.isdir(app_dir):
            print>>sys.stderr, "%s directory doesn't exist." % app_dir
            return
        
        userconf = get_userconf()
        if 'default' in userconf:
            conf = {
                "env": {
                    "default": userconf['env']['default']
                }
            }
        else:
            conf = {}
            
        rc_file = '%s/.couchapprc' % app_dir
        if not os.path.isfile(rc_file):
            if db_url:
                conf.update({ "env": { 
                    "default": {
                        "db": db_url
                    }
                }})

            write_json(rc_file, conf)
        else:
            print>>sys.stderr, "CouchApp already initialized in %s." % app_dir

    def load_metadata(self, app_dir):
        """Reads the .couchapprc to get configuration data"""
        rc_file = os.path.join(app_dir, '.couchapprc')
        conf = get_userconf()
        if os.path.isfile(rc_file):
            conf.update(read_json(rc_file))
        self.conf = conf

    def push_app(self, app_dir, app_name, verbose=False, **kwargs):
        """Pushes the app specified to the CouchDB instance"""
        docid = '_design/%s' % app_name

        attach_dir = os.path.join(app_dir, '_attachments')

        manifest = []
        self.doc = doc = self.dir_to_fields(app_dir, manifest=manifest,
                verbose=verbose)

        self.objects = {}
        if 'shows' in doc:
            self.package_shows(doc['shows'], app_dir, verbose=verbose)

        if 'lists' in doc:
            self.package_shows(doc['lists'], app_dir, verbose=verbose)

        if 'views' in doc:
            self.package_views(doc["views"], app_dir, verbose=verbose)


        for db in self.db:
            if verbose >= 1:
                print "Pushing CouchApp in %s to design doc:\n%s/_design/%s" % (app_dir,
                    db.resource.uri, app_name)
                couchapp = doc.get('couchapp', False)
                if couchapp:
                  index = couchapp.get('index', False)
                  if index:
                    index_url = self.make_index_url(db.resource.uri, app_name, index)
                    print "Visit your CouchApp here:\n%s" % index_url

            new_doc = doc.copy()

            if docid in db:
                design = db[docid]

                _app_meta = design.get('couchapp', {})

                app_meta = {
                    'manifest': manifest,
                    'signatures': _app_meta.get('signatures', {}),
                    'objects': self.objects
                }

                new_doc.update({
                    '_id': docid,
                    '_rev': design['_rev'],
                    'couchapp': app_meta,
                    '_attachments': design.get('_attachments', {})
                })
            else:
                new_doc.update({
                    'couchapp': {
                        'manifest': manifest,
                        'objects': self.objects
                    }
                })

            if 'couchapp' in doc:
                new_doc['couchapp'].update(doc['couchapp'])

            db[docid] = new_doc 

        if 'css' in new_doc['couchapp']:
            if not kwargs.get('nocss', False):
                # merge and compress css
                self.merge_css(attach_dir, doc['couchapp']['css'],
                    docid, verbose=verbose)

        if 'js' in new_doc['couchapp']:
            if not kwargs.get('nojs', False):
                # merge and compress js
                self.merge_js(attach_dir, doc['couchapp']['js'],
                    docid, verbose=verbose)

                    
        self.push_directory(attach_dir, docid, verbose=verbose)
        self.vendor_attachments(app_dir, docid, verbose=verbose)
        

    @classmethod
    def clone(cls, app_uri, app_dir, verbose=False):
        """Clone a CouchApp from app_uri into app_dir"""
        server_uri, db_name, docid = parse_uri(app_uri) 
        couchdb_server = _server(server_uri)

        try:
            db = couchdb_server.create(db_name)
        except: # db already exist
            db = couchdb_server[db_name]
 
        app_name = get_appname(docid)
        if verbose >= 1:
            print "Cloning %s to %s..." % (app_name, app_dir)
        if not app_dir:
            app_dir = os.path.normpath(os.path.join(os.getcwd(), app_name))

        rc_file = os.path.join(app_dir, '.couchapprc')

        if not os.path.isdir(app_dir):
            os.makedirs(app_dir)
        else:
            # delete only if there is .couchapp folder
            if os.path.isfile(rc_file):
                for root, dirs, files in os.walk(app_dir,
                        topdown=False):
                    if root == app_dir:
                        if '_attachments' in dirs:
                            dirs.remove('_attachments') 
                        if '.couchapprc' in files:
                            files.remove('.couchapprc')
                    for name in files:
                        os.remove(os.path.join(root, name))

                    for name in dirs:
                        os.rmdir(os.path.join(root, name))
        
        try:
            design = db[docid]
        except ResourceNotFound:
            print >>sys.stderr, "%s don't exist" % app_name
            return

        metadata = design.get('couchapp', {})
        
        # get manifest
        manifest = metadata.get('manifest', {})

        # get signatures
        signatures = metadata.get('signatures', {})

        # get objects refs
        objects = metadata.get('objects', {})

        conf = read_json(rc_file)
        if not 'env' in conf:
            conf['env'] = {}
        conf['env'].update({
            'origin': {
                'db': db.resource.uri
            }
        })

        write_json(rc_file, conf) 

        # create files from manifest
        if manifest:
            for filename in manifest:
                if verbose >=2:
                    print "clone property: %s" % filename
                file_path = os.path.join(app_dir, filename)
                if filename.endswith('/'): 
                    if not os.path.isdir(file_path):
                        os.makedirs(file_path)
                elif filename == "couchapp.json":
                    continue
                else:
                    parts = filename.split('/')
                    fname = parts.pop()
                    v = design
                    while 1:
                        try:
                            for key in parts:
                                v = v[key]
                        except KeyError:
                            break

                        # remove extension
                        last_key, ext = os.path.splitext(fname)

                        # make sure key exist
                        try:
                            content = v[last_key]
                        except KeyError:
                            break

                        if isinstance(content, basestring):
                            _ref = _md5(to_bytestring(content)).hexdigest()
                            if objects and _ref in objects:
                                content = objects[_ref]

                        if fname.endswith('.json'):
                            content = json.dumps(content)

                        del v[last_key]

                        # make sure file dir have been created
                        file_dir = os.path.dirname(file_path)
                        if not os.path.isdir(file_dir):
                            os.makedirs(file_dir)
                        
                        write_content(file_path, content)

                        # remove the key from design doc
                        temp = design
                        for key2 in parts:
                            if key2 == key:
                                if not temp[key2]:
                                    del temp[key2]
                                break
                            temp = temp[key2]
        
        # second pass for missing key or in case
        # manifest isn't in app
        for key in design.iterkeys():
            if key.startswith('_'): 
                continue
            elif key in ('couchapp'):
                app_meta = design['couchapp'].copy()
                if 'signatures' in app_meta:
                    del app_meta['signatures']
                if 'manifest' in app_meta:
                    del app_meta['manifest']
                if 'objects' in app_meta:
                    del app_meta['objects']
                if app_meta:
                    couchapp_file = os.path.join(app_dir, 'couchapp.json')
                    write_json(couchapp_file, app_meta)
            elif key in ('views'):
                vs_dir = os.path.join(app_dir, key)
                if not os.path.isdir(vs_dir):
                    os.makedirs(vs_dir)
                for vsname, vs_item in design[key].iteritems():
                    vs_item_dir = os.path.join(vs_dir, vsname)
                    if not os.path.isdir(vs_item_dir):
                        os.makedirs(vs_item_dir)
                    for func_name, func in vs_item.iteritems():
                        filename = os.path.join(vs_item_dir, '%s.js' % 
                                func_name)
                        open(filename, 'w').write(func)
                        if verbose >=2:
                            print "clone view not in manifest: %s" % filename
            elif key in ('shows', 'lists'):
                dir = os.path.join(app_dir, key)
                if not os.path.isdir(dir):
                    os.makedirs(dir)
                for func_name, func in design[key].iteritems():
                    filename = os.path.join(dir, '%s.js' % 
                            func_name)
                    open(filename, 'w').write(func)
                    if verbose >=2:
                        print "clone show or list not in manifest: %s" % filename
            else:
                file_dir = os.path.join(app_dir, key)
                if verbose >=2:
                    print "clone property not in manifest: %s" % key
                if isinstance(design[key], (list, tuple,)):
                    write_json(file_dir + ".json", design[key])
                elif isinstance(design[key], dict):
                    if not os.path.isdir(file_dir):
                        os.makedirs(file_dir)
                    for field, value in design[key].iteritems():
                        field_path = os.path.join(file_dir, field)
                        if isinstance(value, basestring):
                            write_content(field_path, value)
                        else:
                            write_json(field_path + '.json', value)        
                else:
                    value = design[key]
                    if not isinstance(value, basestring):
                        value = str(value)
                    write_content(file_dir, value)
   

        # get attachments
        if '_attachments' in design:
            attach_dir = os.path.join(app_dir, '_attachments')
            if not os.path.isdir(attach_dir):
                os.makedirs(attach_dir)
            for filename in design['_attachments'].iterkeys():
                if filename.startswith('vendor'):
                    attach_parts = filename.split('/')
                    vendor_attach_dir = os.path.join(app_dir, attach_parts.pop(0),
                            attach_parts.pop(0), '_attachments')
                    file_path = os.path.join(vendor_attach_dir, '/'.join(attach_parts))
                else:
                    file_path = os.path.join(attach_dir, filename)
                current_dir = os.path.dirname(file_path)
                if not os.path.isdir(current_dir):
                    os.makedirs(current_dir)
        
                if signatures.get(filename) != sign_file(file_path):
                    content = db.get_attachment(docid, filename)
                    write_content(file_path, content)
                    if verbose>=2:
                        print "clone attachment: %s" % filename

    def dir_to_fields(self, app_dir, current_dir='', depth=0,
            manifest=[], verbose=False):
        fields={}
        if not current_dir:
            current_dir = app_dir
        for name in os.listdir(current_dir):
            current_path = os.path.join(current_dir, name)
            rel_path = current_path.split("%s/" % app_dir)[1]
            if name.startswith('.'):
                continue
            elif name.startswith('_'):
                # files starting with "_" are always "special"
                continue
            elif depth == 0 and name in ('couchapp', 'couchapp.json'):
                # we are in app_meta
                if name == "couchapp":
                    manifest.append('%s/' % rel_path)
                    content = self.dir_to_fields(app_dir, current_path,
                        depth=depth+1, manifest=manifest)
                else:
                    manifest.append(rel_path)
                    content = read_json(current_path)
                    if not isinstance(content, dict):
                        content = { "meta": content }
                if 'signatures' in content:
                    del content['signatures']

                if 'manifest' in content:
                    del content['manifest']

                if 'objects' in content:
                    del content['objects']

                if 'couchapp' in fields:
                    fields['couchapp'].update(content)
                else:
                    fields['couchapp'] = content
            elif os.path.isdir(current_path):
                manifest.append('%s/' % rel_path)
                fields[name] = self.dir_to_fields(app_dir, current_path,
                        depth=depth+1, manifest=manifest,
                        verbose=verbose)
            else:
                if verbose >= 2:
                    print >>sys.stderr, "push %s" % rel_path                
                content = ''
                try:
                    content = read_file(current_path)
                except UnicodeDecodeError, e:
                    print >>sys.stderr, str(e)
                if name.endswith('.json'):
                    try:
                        content = json.loads(content)
                    except ValueError:
                        if verbose >= 2:
                            print >>sys.stderr, "Json invalid in %s" % current_path
                
                # remove extension
                name, ext = os.path.splitext(name)
                if name in fields:
                    if verbose >= 2:
                        print >>sys.stderr, "%(name)s is already in properties. Can't add (%(name)s%(ext)s)" % {
                        "name": name,
                        "ext": ext
                        }
                else:
                    manifest.append(rel_path)
                    fields[name] = content
        return fields
    
    def _put_attachment(self, db, doc, content, filename):
        nb_try = 0
        while True:
            error = False
            try:
                db.put_attachment(doc, content, filename)
            except (socket.error, httplib.BadStatusLine):
                time.sleep(0.4)
                error = True

            nb_try = nb_try +1
            if not error:
                break

            if nb_try > 3:
                if verbose >= 2:
                    print >>sys.stderr, "%s file not uploaded, sorry." % filename
                break
                
    def vendor_attachments(self, app_dir, docid, verbose):
        vendor_dir = os.path.join(app_dir, 'vendor')
        if not os.path.isdir(vendor_dir):
            return
            
        for name in os.listdir(vendor_dir):
            current_path = os.path.join(vendor_dir, name)
            if os.path.isdir(current_path):
                attach_dir = os.path.join(current_path, '_attachments')
                if os.path.isdir(attach_dir):
                    self.push_directory(attach_dir, docid, verbose, 
                                    vendor=name)
                    
    def push_directory(self, attach_dir, docid, verbose=False, vendor=None):
        # get attachments
        _signatures = {}
        _attachments = {}
        all_signatures = {}
        for root, dirs, files in os.walk(attach_dir):
            if files:
                for filename in files:
                    if filename.startswith('.'):
                        continue
                    else:
                        file_path = os.path.join(root, filename) 
                        name = file_path.split('%s/' % attach_dir)[1]
                        if vendor is not None:
                            name = os.path.join('vendor/%s' % vendor, name)
                        signature = sign_file(file_path)
                        _signatures[name] = signature
                        _attachments[name] = open(file_path, 'rb')
        
        # detect attachments to be removed and keep
        # only new version attachments to update.
        for db in self.db:
            design = db[docid]
            metadata = design.get('couchapp', {})
            attachments = _attachments.copy()
            if 'signatures' in metadata:
                all_signatures = metadata['signatures'].copy()
                for filename in metadata['signatures'].iterkeys():
                    if vendor is not None:
                        if filename.startswith('vendor/%s' % vendor):
                            del all_signatures[filename]
                            if filename not in _signatures:
                                db.delete_attachment(design, filename)
                            elif _signatures[filename] == metadata['signatures'][filename]:
                                del attachments[filename]
                            
                    else:
                        if not filename.startswith('vendor'):
                            del all_signatures[filename]
                            if filename not in _signatures:
                                db.delete_attachment(design, filename)
                            else:
                                if _signatures[filename] == metadata['signatures'][filename]:
                                    del attachments[filename]

            for filename, value in attachments.iteritems():
                if verbose >= 2:
                    print "Attaching %s" % filename
               
                # fix issue with httplib that raises BadStatusLine
                # error because it didn't close the connection
                self._put_attachment(db, design, value, filename)
                         
            # update signatures
            design = db[docid]
            if not 'couchapp' in design:
                design['couchapp'] = {}

            all_signatures.update(_signatures)
                
            design['couchapp'].update({'signatures': all_signatures})
            db[docid] = design

    def package_shows(self, funcs, app_dir, verbose=False):
        self.apply_lib(funcs, app_dir, verbose=verbose)
              
    def package_views(self, views, app_dir, verbose=False):
        for view, funcs in views.iteritems():
            self.apply_lib(funcs, app_dir, verbose=verbose)


    def apply_lib(self, funcs, app_dir, verbose=False):
        if not hasattr(self, "objects"):
            self.objects = {}
        for k, v in funcs.iteritems():
            if not isinstance(v, basestring):
                continue
            old_v = v
            try:
              funcs[k] = self.run_json_macros(
                              self.run_code_macros(v, app_dir, verbose=verbose), 
                              app_dir, verbose=verbose)
            except ValueError, e:
              print >>sys.stderr, "Error running !code or !json on function \"%s\": %s" % (k, e)
              sys.exit(-1)
            if old_v != funcs[k]:
                self.objects[_md5(to_bytestring(funcs[k])).hexdigest()] = old_v

    def run_code_macros(self, f_string, app_dir, verbose=False):
        def rreq(mo):
            # just read the file and return it
            path = os.path.join(app_dir, mo.group(2).strip(' '))
            library = ''
            filenum = 0
            for filename in glob.iglob(path):            
                if verbose>=2:
                    print "process code macro: %s" % filename
                try:
                    library += read_file(filename)
                except IOError, e:
                    print >>sys.stderr, e
                    sys.exit(-1)
                filenum += 1
                
            if not filenum:
                print >>sys.stderr, "Processing code: No file matching '%s'" % mo.group(2)
                sys.exit(-1)
                
            return library

        re_code = re.compile('(\/\/|#)\ ?!code (.*)')
        return re_code.sub(rreq, f_string)

    def run_json_macros(self, f_string, app_dir, verbose=False):
        included = {}
        varstrings = []

        def rjson(mo):
            if mo.group(2).startswith('_attachments'): 
                # someone  want to include from attachments
                path = os.path.join(app_dir, mo.group(2).strip(' '))
                filenum = 0
                for filename in glob.iglob(path):
                    library = ''
                    try:
                        if filename.endswith('.json'):
                            library = read_json(filename)
                        else:
                            library = read_file(filename)
                    except IOError, e:
                        print >>sys.stderr, e
                        sys.exit(1)
                    filenum += 1
                    current_file = filename.split(app_dir)[1]
                    fields = current_file.split('/')
                    count = len(fields)
                    include_to = included
                    for i, field in enumerate(fields):
                        if i+1 < count:
                            include_to[field] = {}
                            include_to = include_to[field]
                        else:
                            include_to[field] = library
                if not filenum:
                    print >>sys.stderr, "Processing code: No file matching '%s'" % mo.group(2)
                    sys.exit(-1)
            else:	
                fields = mo.group(2).split('.')
                library = self.doc
                count = len(fields)
                include_to = included
                for i, field in enumerate(fields):
                    if not field in library: break
                    library = library[field]
                    if i+1 < count:
                        include_to[field] = include_to.get(field, {})
                        include_to = include_to[field]
                    else:
                        include_to[field] = library

            return f_string

        def rjson2(mo):
            return '\n'.join(varstrings)

        re_json = re.compile('(\/\/|#)\ ?!json (.*)')
        re_json.sub(rjson, f_string)

        if not included:
            return f_string

        for k, v in included.iteritems():
            varstrings.append("var %s = %s;" % (k, json.dumps(v)))

        return re_json.sub(rjson2, f_string)

    def merge_css(self, attach_dir, css_conf, docid, verbose=False):
        re_url = re.compile('url\s*\(([^\s"].*)\)')

        src_fpath = ''
        fname_dir = ''

        def replace_url(mo):
            """ make sure urls are relative to css path """
            css_url = mo.group(0)[4:].strip(")").replace("'", "").replace('"','')
            css_path = os.path.join(os.path.dirname(src_fpath),
                    css_url)

            rel_path = _relpath(css_path, fname_dir)
            return "url(%s)" % rel_path
        
        for fname, src_files in css_conf.iteritems():
            output_css = ''

            dest_path = os.path.join(attach_dir, fname)
            fname_dir = os.path.dirname(dest_path)

            for src_fname in src_files:
                src_fpath = os.path.join(attach_dir, src_fname)
                
                if os.path.exists(src_fpath):
                    content_css = str(CSSParser(read_file(src_fpath)))
                    content_css = re_url.sub(replace_url, content_css) 
                    output_css += content_css
                    if verbose >= 2:
                        print "Merging %s in %s" % (src_fname, fname)

            if not os.path.isdir(fname_dir):
                os.makedirs(fname_dir)

            write_content(dest_path, output_css) 
            
    def make_index_url(self, uri, app_name, index):
        return "%s/%s/%s/%s" % (uri, index[0], app_name, index[1])
    
    @classmethod
    def vendor_update(cls, app_dir, verbose=False):
        vendor_dir = os.path.join(app_dir, "vendor")
        if not os.path.isdir(vendor_dir):
            return

        vendor_handlers = _vendor_handlers()
 
        for name in os.listdir(vendor_dir):
            current_path = os.path.join(vendor_dir, name)
            if os.path.isdir(current_path):
                mfile = os.path.join(current_path, 'metadata.json')
                metadata = read_json(mfile)
                if not metadata and name == 'couchapp':
                    update_url = COUCHAPP_VENDOR_URL
                    scm = COUCHAPP_VENDOR_SCM
                elif metadata:
                    update_url = metadata['update_url']
                    scm = metadata['scm']
                    if not scm in VENDOR_HANDLERS:
                        scm = False
                
                if update_url and scm:
                    # for now we manage only internal handlers
                    handler = vendor_handlers[scm]
                    cmd = "%s update %s %s %s" % (handler, update_url, 
                                            current_path, vendor_dir)
                    (child_stdin, child_stdout, child_stderr) = _popen3(cmd)
                    if verbose >=2:
                        print child_stdout.read()
                        err = child_stderr.read()
                        if err:
                            print >>sys.stderr, err
                        
    @classmethod
    def vendor_install(cls, app_dir, url, scm="git", verbose=False):   
        if not scm in VENDOR_HANDLERS:
            print >>sys.stderr, "%s scm isn't supported yet." % scm
            sys.exit(-1)
        vendor_dir = os.path.join(app_dir, "vendor")
        if not os.path.isdir(vendor_dir):
            os.makedirs(vendor_dir)
            
        vendor_handlers = _vendor_handlers()
        
        # get list of installed applications
        installed_apps = []
        for name in os.listdir(vendor_dir):
            current_path = os.path.join(vendor_dir, name)
            if os.path.isdir(current_path):
                installed_apps.append(name)
                
            
        handler = vendor_handlers[scm]
        cmd = "%s install %s %s" % (handler, url, vendor_dir)
        (child_stdin, child_stdout, child_stderr) = _popen3(cmd)
        err = child_stderr.read()
        if verbose >=2:
            print child_stdout.read()
        if err:
            print >>sys.stderr, err
                
        # detect new vendor application and add url so we could update later
        for name in os.listdir(vendor_dir):
            current_path = os.path.join(vendor_dir, name)
            if os.path.isdir(current_path) and name not in installed_apps:
                new_file = os.path.join(current_path, '.new')
                if os.path.isfile(new_file):
                    new_url = read_file(new_file).strip()
                    if new_url == url:
                        mfile = os.path.join(current_path, 'metadata.json')
                        write_json(mfile, {
                            "scm": scm,
                            "update_url": url
                        })
                        os.unlink(new_file)
                        return
                
    def merge_js(self, attach_dir, js_conf, docid, verbose=False):
        if "js_compressor" in self.conf:
            if not isinstance(self.conf["js_compressor"], basestring):
                print >>sys.stderr, "Warning: js_compressor settings should be a string"
                print >>sys.stderr, "         Selecting default backend (jsmin)"
                import couchapp.utils.jsmin as backend
            else:
                try:
                    backend = __import__(self.conf['js_compressor'], {}, {}, [''])
                except ImportError:
                    import couchapp.utils.jsmin as backend
        else:
            import couchapp.utils.jsmin as backend

        if verbose:
            backend.about()

        for fname, src_files in js_conf.iteritems():
            output_js = ''

            dest_path = os.path.join(attach_dir, fname)
            fname_dir = os.path.dirname(dest_path)

            for src_fname in src_files:
                src_fpath = os.path.join(attach_dir, src_fname)
                if os.path.isfile(src_fpath):
                    output_js += "/* %s */\n" % src_fpath
                    output_js +=  read_file(src_fpath)
                    if verbose >= 2:
                        print "merging %s in %s" % (src_fname, fname)

            if not os.path.isdir(fname_dir):
                os.makedirs(fname_dir)

            output_js = backend.compress(output_js)
            write_content(dest_path, output_js) 




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.
#

import codecs
import os
import sys
import urlparse
import urllib

# compatibility with python 2.4
try:
    from hashlib import md5 as _md5
except ImportError:
    import md5
    _md5 = md5.new

try:
    import json 
except ImportError:
    import simplejson as json 


def in_couchapp():
    current_path = os.getcwd()
    old_dirs = []
    while 1:
        dirs = os.listdir(current_path)
        if dirs == old_dirs: 
            return False
        if '.couchapprc' in dirs: break
        current_path = os.path.normpath(os.path.join(current_path, '../'))
        old_dirs = dirs
    return current_path

def parse_uri(string):
    parts = urlparse.urlsplit(urllib.unquote(string))
    if parts[0] != 'http' and parts[0] != 'https':
        raise ValueError('Invalid dbstring')
     
    path = parts[2].strip('/').split('/')

    dbname = ''
    docid = ''
    if len(path) >= 1:
        db_parts=[]
        i = 0
        while 1:
            try:
                p = path[i]
            except IndexError:
                break

            if p == '_design': break
            db_parts.append(p)
            i = i + 1
        dbname = '/'.join(db_parts)
        
        if i < len(path) - 1:
            docid = '/'.join(path[i:])

    server_uri = '%s://%s' % (parts[0], parts[1])
    return server_uri, dbname, docid


def parse_auth(string):
    parts = urlparse.urlsplit(urllib.unquote(string))
    
    server_parts = parts[1].split('@')
    if ":" in server_parts[0]:
        username, password = server_parts[0].split(":")
    else:
        username = server_parts[0]
        password = ''

    server_uri = "%s://%s" % (parts[0], server_parts[1])

    return username, password, server_uri

def get_appname(docid):
    return docid.split('_design/')[1]


def to_bytestring(s):
    if not isinstance(s, basestring):
        return s
    if isinstance(s, unicode):
        return s.encode('utf-8')
    else:
        return s

def read_file(fname):
    f = codecs.open(fname, 'rb', "utf-8")
    data = f.read()
    f.close()
    return data

def sign_file(file_path):
    if os.path.isfile(file_path):
        f = open(file_path, 'rb')
        content = f.read()
        f.close()
        return _md5(content).hexdigest()
    return ''

def write_content(fname, content):
    f = open(fname, 'wb')
    f.write(to_bytestring(content))
    f.close()

def write_json(filename, content):
    write_content(filename, json.dumps(content))

def read_json(filename):
    try:
        data = read_file(filename)
    except IOError, e:
        if e[0] == 2:
            return {}
        raise
    try:
        data = json.loads(data)
    except ValueError:
        print >>sys.stderr, "Json is invalid, can't load %s" % filename
        return {}
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.

try:
    __version__ = __import__('pkg_resources').get_distribution('Couchapp').version
except:
    __version__ = '?'

from couchapp.file_manager import FileManager



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages
from setuptools.command.easy_install import easy_install

import os
import sys

data_files = []

for dir, dirs, files in os.walk('app-template'):
    data_files.append((os.path.join('couchapp', dir), 
        [os.path.join(dir, file_) for file_ in files]))

for dir, dirs, files in os.walk('vendor'):
    data_files.append((os.path.join('couchapp', dir), 
        [os.path.join(dir, file_) for file_ in files]))

for dir, dirs, files in os.walk('python/couchapp'):
    for i, dirname in enumerate(dirs):
        if dirname.startswith('.'): del dirs[i]
        
    data_files.append((dir, [os.path.join(dir, file_) for file_ in files]))
    
easy_install.real_process_distribution = easy_install.process_distribution
def process_distribution(self, *args, **kwargs):
    """ overide process_distribution to add permissions"""
    easy_install.real_process_distribution(self, *args, **kwargs)
    import pkg_resources
    external_path = '/pathto/couchapp/_external'
    try:
        pkg_resources.require('couchapp')
        external_path = pkg_resources.resource_filename("couchapp", "_external")
        for dir, dirs, files in os.walk(external_path):
            for i, dirname in enumerate(dirs):
                if dirname.startswith('.'): del dirs[i]
            for file_ in files:
                os.chmod(os.path.join(dir, file_), 0755)  
            
    except:
        print >>sys.stderr, "Chmoding failed. Try to 'chmod -R +x %s'" % external_path
easy_install.process_distribution = process_distribution        
 
setup(
    name = 'Couchapp',
    version = '0.1.14',
    url = 'http://github.com/benoitc/couchapp/tree/master',
    license =  'Apache License 2',
    author = 'Benoit Chesneau',
    author_email = 'benoitc@e-engura.org',
    description = 'Standalone CouchDB Application Development Made Simple.',
    long_description = """CouchApp is a set of helpers and a jQuery plugin
    that conspire to get you up and running on CouchDB quickly and
    correctly. It brings clarity and order to the freedom of CouchDB's
    document-based approach.""",
    keywords = 'couchdb couchapp',
    platforms = ['any'],

    zip_safe = False,
    
    packages=find_packages('python'),
    package_dir={
        '': 'python'
    },
    data_files = data_files,
    include_package_data = True,
    entry_points = {
        'console_scripts': [
            'couchapp = couchapp.bin.couchapp_cli:main',
        ]
    },
    classifiers = [
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Topic :: Database',
        'Topic :: Utilities',
    ],

    setup_requires = [
        'setuptools>=0.6c7',
    ],

    install_requires = [
        'couchdb>=0.5',
        'simplejson',
    ],
    
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.
#


"""
simple backend to use yuicompressor to compress files
"""

import codecs
import os
from popen2 import popen2
import tempfile

from couchapp.utils import read_file


def about():
    print "Compressing javascript with yuicompressor backend"

def compress(js):
    cmd_path = os.path.join(os.path.dirname(__file__), 
            'yuicompressor-2.4.1.jar')
    fd, fname = tempfile.mkstemp()
    f = codecs.getwriter('utf8')(os.fdopen(fd, "w"))
    f.write(js)
    f.close()
    cmd = "java -jar %s --type js %s" % (cmd_path, fname) 
    sout, sin = popen2(cmd)
    data = sout.read()
    os.unlink(fname) 
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2008 Jan Lehnardt <jan@apache.org>
# Copyright 2009 Benoit Chesneau <benoitc@e-engura.org>
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution.

"""
Script that pushes a file system hierarchy into a CouchDB design document.

Good for pure Couch application development.

A port of couchapp from Ruby (http://github.com/jchris/couchapp)
"""

import os
import sys
from optparse import OptionParser, OptionGroup


import couchapp
from couchapp.utils import in_couchapp


def generate(appname, verbose=False):
    appdir = os.path.normpath(os.path.join(os.getcwd(), appname))
    if verbose >= 1:
        print "Generating a new CouchApp in %s" % appdir
    couchapp.FileManager.generate_app(appdir)

def init(appdir, dburl, verbose=False):
    if verbose >= 1:
        print "Initializing a new CouchApp in %s" % appdir
    couchapp.FileManager.init(appdir, dburl)

def push(appdir, appname, dbstring, verbose=False, 
        options=None):
    try:
        fm = couchapp.FileManager(dbstring, appdir)
    except ValueError, e:
        print>>sys.stderr, e
        return 

    fm.push_app(appdir, appname, verbose=verbose, 
            nocss=options.css, nojs=options.js)

def clone(app_uri, app_dir, verbose=False):
    couchapp.FileManager.clone(app_uri, app_dir, verbose=verbose)
    
def vendor_update(app_dir, verbose=False):
    couchapp.FileManager.vendor_update(app_dir, verbose=verbose)
    
def vendor_install(app_dir, url, scm='git', verbose=False):
    couchapp.FileManager.vendor_install(app_dir, url, scm=scm, 
                                    verbose=verbose)

def main():
    parser = OptionParser(usage='%prog [options] cmd', version="%prog " + couchapp.__version__)
    parser.add_option('-v', dest='verbose', default=1, help='print message to stdout')
    parser.add_option('-q', dest='verbose', action='store_const', const=0, help="don't print any message")
    
    # generate options
    parser.add_option_group(OptionGroup(parser, "Generate a new CouchApp! (start here)",
            "couchapp generate <appname> [appdir]"))
            
    # push options
    group_push = OptionGroup(parser, "Pushes a CouchApp to CouchDB", 
            "couchapp push [options] [appdir] [appname] [dburl]")
    group_push.add_option("--disable-css", action="store_true", 
        dest="css", help="disable css compression")
    group_push.add_option("--disable-js", action="store_true", 
        dest="js", help="disable js compression")
    parser.add_option_group(group_push)
    
    # clone options
    parser.add_option_group(OptionGroup(parser, 
        "Clones/Pulls a CouchApp from a url (like http://host/db/_design/CA_name)",
        "couchapp clone/pull <dburl> [dir]"))
            
    # init options
    group_init = OptionGroup(parser, "Initialize CouchApp .couchapprc", "couchapp init [options] <appdir>")
    group_init.add_option("--db", action="store", help="full url of default database")
    parser.add_option_group(group_init)
    
    group_vendor = OptionGroup(parser, "Install a vendor", "couchapp vendor install vendor_url [option][appdir]")
    group_vendor.add_option("--scm", action="store", default='git', 
                                help="scm used to install the vendor, by default git")
    parser.add_option_group(group_vendor)
    
    options, args = parser.parse_args()


    if len(args) < 1:
        return parser.error('incorrect number of arguments')

    if args[0] == 'generate':
        if len(args) < 2:
            return parser.error('cmd: "generate appname"'+
                    '\n\nIncorrect number of arguments, appname is'+
                    ' missing')
        appname = args[1]
        generate(appname, options.verbose)
    elif args[0] == 'push':
        appname = ''
        rel_path = '.'
        dbstring = ''
        # generate [dir] [appname] [url] case
        if len(args) == 4:
            if in_couchapp():
                return parser.error('Incorrect number of arguments, you\'re in an app.')
            rel_path = args[1]
            appname = args[2]
            dbstring = args[3]
            
        # generate [dir/appname] [url] case
        elif len(args) == 3:
            rel_path = in_couchapp()
            if rel_path:
                if args[1] != '.':
                    appname = args[1]
                dbstring = args[2]
            else:
                rel_path = args[1]
                dbstring = args[2]
        
        # generate [dir/url] case
        elif len(args) == 2:
            rel_path = in_couchapp()
            if rel_path:
                if args[1] != '.':
                    dbstring = args[1]
            else:
                rel_path = args[1]
                
        # just generate
        elif len(args) == 1:
            rel_path = in_couchapp()
            if not rel_path:
                rel_path = '.'

        appdir = os.path.normpath(os.path.join(os.getcwd(), rel_path))
        # Derive appname from the directory name (/home/foo/sofa => sofa)
        if not appname: 
            appname = ''.join(appdir.split('/')[-1:])
        # PUSH IT!
        push(appdir, appname, dbstring, options.verbose, options=options)
      
    elif args[0] == 'clone' or args[0] == 'pull':
        if len(args) < 2:
            return parser.error('Incorrect number of arguments (at least two)')
        # clone/pull <url> [dir] case
        if len(args) == 3:
            app_dir = args[2]
        # clone/pull <url>
        else:
            app_dir = ''
        # CLONE IT!
        clone(args[1], app_dir, options.verbose)
    
    elif args[0] == 'init':
        dburl = options.db or ''
        try:
            appdir = args[1]
        except IndexError:
            appdir = '.'
        init(appdir, dburl, options.verbose)
    elif args[0] == 'vendor':
        if len(args) < 2:
            return parser.error('Incorrect number of arguments (at least two)')
       
            
        action = args[1]
        if action == 'update':
            try:
                appdir = args[2]
            except IndexError:
                appdir = '.'
            vendor_update(appdir, options.verbose)
        elif action == 'install':
            if len(args) < 3:
                return parser.error('Incorrect number of arguments')
                
            try:
                appdir = args[3]
            except IndexError:
                appdir = '.'
            vendor_install(appdir, args[2], options.scm, options.verbose)
        else:
            print >>sys.stderr, "%s is an unknown vendor action, sorry." % action
            
        
    else:
        print "%s is an unknown command, sorry." % args[0]

if __name__ == '__main__':
    main()
